/************************************************************************
 *   IRC - Internet Relay Chat, src/m_error.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   $Id: m_error.c,v 1.9 2003/06/24 03:57:16 ievil Exp $
 */
#include "m_commands.h"
#include "client.h"
#include "common.h"   /* FALSE */
#include "ircd.h"
#include "numeric.h"
#include "send.h"
#include "s_debug.h"
#include "s_log.h"

/*
 * m_functions execute protocol messages on this server:
 *
 *      cptr    is always NON-NULL, pointing to a *LOCAL* client
 *              structure (with an open socket connected!). This
 *              identifies the physical socket where the message
 *              originated (or which caused the m_function to be
 *              executed--some m_functions may call others...).
 *
 *      sptr    is the source of the message, defined by the
 *              prefix part of the message if present. If not
 *              or prefix not found, then sptr==cptr.
 *
 *              (!IsServer(cptr)) => (cptr == sptr), because
 *              prefixes are taken *only* from servers...
 *
 *              (IsServer(cptr))
 *                      (sptr == cptr) => the message didn't
 *                      have the prefix.
 *
 *                      (sptr != cptr && IsServer(sptr) means
 *                      the prefix specified servername. (?)
 *
 *                      (sptr != cptr && !IsServer(sptr) means
 *                      that message originated from a remote
 *                      user (not local).
 *
 *              combining
 *
 *              (!IsServer(sptr)) means that, sptr can safely
 *              taken as defining the target structure of the
 *              message in this server.
 *
 *      *Always* true (if 'parse' and others are working correct):
 *
 *      1)      sptr->from == cptr  (note: cptr->from == cptr)
 *
 *      2)      MyConnect(sptr) <=> sptr == cptr (e.g. sptr
 *              *cannot* be a local connection, unless it's
 *              actually cptr!). [MyConnect(x) should probably
 *              be defined as (x == x->from) --msa ]
 *
 *      parc    number of variable parameter strings (if zero,
 *              parv is allowed to be NULL)
 *
 *      parv    a NULL terminated list of parameter pointers,
 *
 *                      parv[0], sender (prefix string), if not present
 *                              this points to an empty string.
 *                      parv[1]...parv[parc-1]
 *                              pointers to additional parameters
 *                      parv[parc] == NULL, *always*
 *
 *              note:   it is guaranteed that parv[0]..parv[parc-1] are all
 *                      non-NULL pointers.
 */

/*
 * Note: At least at protocol level ERROR has only one parameter,
 * although this is called internally from other functions
 * --msa
 *
 *      parv[0] = sender prefix
 *      parv[*] = parameters
 */
int m_error(struct Client *cptr, struct Client *sptr, int parc, char *parv[])
{
  char* para;

  para = (parc > 1 && *parv[1] != '\0') ? parv[1] : "<>";

  Debug((DEBUG_ERROR,"Received ERROR message from %s: %s",
         sptr->name, para));
  /*
   * Ignore error messages generated by normal user clients
   * (because ill-behaving user clients would flood opers
   * screen otherwise). Pass ERROR's from other sources to
   * the local operator...
   */
  if (IsPerson(cptr) || IsUnknown(cptr))
    return 0;

  /* this is less ugly than it was before, if you ask me.
   * at least with this logic, everything's clustered, and
   * there's not a crapload of code duplic.  anyways, the
   * logic is this:
   * if the source is the server it came from,
   *   if HIDE_ERROR_MESSAGES is undefined,
   *     sendto_realops(hey i got an error me love you long time)
   *       [ and hide the remote servers ip if HIDE_SERVERS_IPS ]
   *   always: ilog(blah blah blah, with the real ip)
   * otherwise, if this was sent along from another server:
   *   same logic as above, but messages contain "via x"
   *
   * booooooooing.
   */

  if (cptr == sptr)
  {
#ifndef HIDE_ERROR_MESSAGES
    sendto_realops("ERROR :from %s -- %s",
#if (defined SERVERHIDE) || (defined HIDE_SERVERS_IPS)
               get_client_name(cptr, MASK_IP), para);
#else /* SERVERHIDE || HIDE_SERVERS_IPS */
               get_client_name(cptr, FALSE), para);
#endif /* SERVERHIDE || HIDE_SERVERS_IPS */
#endif /* !HIDE_ERROR_MESSAGES */
    ilog(L_NOTICE, "ERROR from %s -- %s",
               get_client_name(cptr, SHOW_IP), para);
  }
  else
  {
#ifndef HIDE_ERROR_MESSAGES
    sendto_realops("ERROR :from %s via %s -- %s", sptr->name,
#if (defined SERVERHIDE) || (defined HIDE_SERVERS_IPS)
               get_client_name(cptr, MASK_IP), para);
#else /* SERVERHIDE || HIDE_SERVERS_IPS */
               get_client_name(cptr, FALSE), para);
#endif /* SERVERHIDE || HIDE_SERVERS_IPS */
#endif /* !HIDE_ERROR_MESSAGES */
    ilog(L_NOTICE, "ERROR from %s via %s -- %s", sptr->name,
        get_client_name(cptr, SHOW_IP), para);
  }

  return 0;
}

